# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from datetime import datetime, timedelta
from itemadapter import ItemAdapter
from scrapy.exceptions import DropItem
from scrapy import Request
import logging
import json
import time
import os

from elasticsearch import Elasticsearch
from elasticsearch import helpers
from elasticsearch.client import IndicesClient
from elasticsearch.exceptions import RequestError
from elasticsearch.exceptions import ConnectionTimeout
from elasticsearch.exceptions import ConnectionError
from elasticsearch.exceptions import SSLError
from elasticsearch.exceptions import TransportError, ElasticsearchException, NotFoundError
from elasticsearch.helpers.errors import BulkIndexError

from .logger import get_logger

logger = get_logger(__name__, "app.log")


class ElasticSearchPipeline:
    ELASTICSEARCH_BUFFER_LENGTH = 50
    ELASTICSEARCH_INDEX = None
    FORCE = False
    ACTION = None

    def __init__(self):
        self.settings = None
        self.id_key = None
        self.max_retry = 3
        self.es = None
        self.items_buffer = []

    @classmethod
    def init_es_client(cls, settings):
        es_timeout = settings.get("ELASTICSEARCH_TIMEOUT", 60)
        es_servers = settings.get("ELASTICSEARCH_SERVERS")
        es_servers = es_servers if isinstance(es_servers, list) else [es_servers]

        es_settings = dict()
        es_settings["hosts"] = es_servers
        es_settings["timeout"] = es_timeout

        if settings.get("ELASTICSEARCH_USERNAME") and settings.get("ELASTICSEARCH_PASSWORD"):
            es_settings["http_auth"] = (
                settings["ELASTICSEARCH_USERNAME"],
                settings["ELASTICSEARCH_PASSWORD"],
            )

        return Elasticsearch(**es_settings)

    @classmethod
    def from_crawler(cls, crawler):
        ext = cls()
        ext.settings = crawler.settings

        ext.es = cls.init_es_client(crawler.settings)
        ext.max_retry = crawler.settings.get('ELASTICSEARCH_MAX_RETRY', 3)
        ext.id_key = crawler.settings.get('ELASTICSEARCH_ID_KEY')

        return ext

    def save_item(self, item, action="create"):
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": action,
            "_source": dict(item),
        }

        if '_id' in item and item['_id']:
            index_action['_id'] = item.pop('_id')
        else:
            if self.id_key is not None and self.id_key in item:
                index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def send_items(self):
        retry = self.max_retry
        while retry > 0:
            try:
                res = helpers.bulk(self.es, self.items_buffer)
                logging.info(f"[Items Saved]: {self.items_buffer}")
                break
            except (ConnectionTimeout, ConnectionError, SSLError, TransportError):
                time.sleep(retry)
                retry -= 1
                continue
            except RequestError as e:
                break
            except BulkIndexError as e:
                for error in e.errors:
                    for _, value in error.items():
                        try:
                            logging.exception(value["_id"])
                        except Exception as e:
                            pass
                break
            except Exception as e:
                raise e

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            if self.FORCE:
                self.save_item(item, "index")
            elif self.ACTION:
                self.save_item(item, self.ACTION)
            else:
                self.save_item(item)

            logging.debug("Item has been sent to Elastic Search %s", self.ELASTICSEARCH_INDEX)

        return item

    def ensure_indice(self, indice_name):
        ic = IndicesClient(self.es)
        if ic.exists(indice_name):
            return True

        ic.create(indice_name, {"settings": {"index.mapping.total_fields.limit": 1000}})
        return ic.exists(indice_name)

    def close_spider(self, spider):
        if len(self.items_buffer):
            self.send_items()
